'''
Environment for DNA Mutations
'''
import torch 
import warnings
import numpy as np
import pandas as pd
from gymnasium import Env, spaces
from grelu.lightning import LightningModel

from utils.constants import REFSEQ, ALPHABET, IDXTONUC, seq_to_one_hot
from utils.eval_utils import distance
from net.buffers import * 
# from net.rew import BaseCNN
#from config import * 

warnings.filterwarnings('ignore')
import sys
sys.path.append('/home/v-zhaoyang2/project/DNARL')
# import src.reglm as reglm
import src.reglm.regression 

class SingleOpt(Env):
    def __init__(self, config, seed=42):
        super().__init__()
        
        obs_shape = (config.length, 4) # 4 for DNA
        self.observation_space = spaces.Box(low=0, high=1, shape=obs_shape, dtype=np.float32)
        self.action_space = spaces.MultiDiscrete([4, config.length])
        
        self.device = config.device
        self.cell = config.name
        self.length = config.length
        self.done_cond = config.done_cond
        self.oracle_calls = 0
        self.config = config

        self.oracle = src.reglm.regression.EnformerModel.load_from_checkpoint(config.rew_pretrained, map_location=self.device)
        self.oracle.eval()
        
        # self.oracle.eval()

        data = pd.read_csv('data/{}/{}.csv'.format(config.name, config.level))[["sequence", 'target']]
        data["target"] = (data["target"] - self.config.min_fitness)/self.config.max_fitness
        self.buffer = Buffer(list(data.itertuples(index=False, name=None)), random=random.Random(seed))
        self.inits = data["sequence"].tolist()
                
        self.ep = 0
        self.steps = 0
        self.total_steps = 0
        self.state = None 
        self.state_seq = None # String 
        self.wt_seq = REFSEQ[self.cell][self.config.level]
        self.init_target = 0
        self.buffer_idx = None

        # logging
        self.reward = 0
        self.done = False
        self.target = 0
        self.best_discovered = 0
        self.init_seq = None
        self.n_mut = 0 
        self.nuc = {nuc:0 for nuc in ALPHABET}
        self.pos = {pos:0 for pos in range(self.length)}

    def normalize_target(self, target):
        # return (target - self.config.min_fitness)/(self.config.max_fitness - self.config.min_fitness)
        return target

    def reset(self, seed=422):
        self.state_seq, self.buffer_idx = self.buffer.top()
        self.init_seq = self.state_seq
        self.state = seq_to_one_hot(self.state_seq).numpy()
        with torch.no_grad():
            # target = self.oracle(seq_to_one_hot(self.state_seq).T.unsqueeze(0).to(self.device))[:,0,0]
            # self.init_target = self.target = target = self.normalize_target(target.cpu().item())
            self.init_target = self.oracle([self.init_seq]).detach().cpu().numpy().astype(float).reshape(-1)
        self.ep += 1
        self.steps = 0
        self.total_steps += 1
        return self.state, {}
    
    def step(self, action):
        self.steps += 1
        self.total_steps += 1
        
        nuc = int(action[0])
        pos = int(action[1])
        
        self.nuc[IDXTONUC[nuc]] += 1
        self.pos[pos] += 1
        
        if self.state_seq[pos] != IDXTONUC[nuc]:
            self.step_mut = step_mut = 1
        else:
            self.step_mut = step_mut = 0
        
        next_seq = self.state_seq[:pos] + IDXTONUC[nuc] + self.state_seq[pos+1:]
        self.n_mut = distance(self.wt_seq, next_seq)
        next_state = seq_to_one_hot(next_seq).numpy()
            
        self.done = done = step_mut == 0 or self.steps > self.done_cond.max_steps or self.n_mut > self.done_cond.max_mutation 
        called = False
        
        if self.config.not_sparse or done:
            self.oracle_calls += 1
            with torch.no_grad():
                # _, target = self.oracle(seq_to_one_hot(next_seq).unsqueeze(0).to(self.device), get_embed=True)
                # target = self.oracle(seq_to_one_hot(next_seq).T.unsqueeze(0).to(self.device))[:,0,0]
                target = self.oracle([next_seq]).detach().cpu().numpy().astype(float).reshape(-1)
            target = self.normalize_target(target.item())
            self.reward = target
            self.buffer.push((next_seq, target, self.buffer_idx))
            if target > self.best_discovered:
                self.best_discovered = target
            self.target = target
            called = True
        else:
            self.reward = 0
        
        self.state = next_state
        self.state_seq = next_seq
            
        if done:
            info = {
                'candidates': self.state_seq, 
                'fitness': self.target,
                'init_seq': self.init_seq,
                'n_mut': self.n_mut,
                'nuc': self.nuc,
                'pos': self.pos,
                'called': called
            }
            return self.state, self.reward, done, False, info
        return self.state, self.reward, done, False, {}